@@ -25,16 +25,21 @@ from logging import getLogger
 from cPickle import dumps, loads
 
 from cubicweb import neg_role, validation_error, Binary, ValidationError
+from cubicweb.__pkginfo__ import numversion
 from cubicweb.rset import ResultSet
-from cubicweb.predicates import is_instance
 from cubicweb.schema import RQLConstraint
+from cubicweb import server
 from cubicweb.server.edition import EditedEntity
 from cubicweb.server.utils import eschema_eid
 from cubicweb.server.session import Session
 from cubicweb.hooks.integrity import DONT_CHECK_RTYPES_ON_ADD
-from cubicweb.view import EntityAdapter
+
+from cubes.worker.entities import Performer
 
 from cubes.fastimport.hooks import HooksRunner
+from cubes.fastimport import utils # monkeypatch the native source
+
+notcw319 = numversion[:2] < (3, 19)
 
 
 YAMS_TO_PY_TYPEMAP = defaultdict(
@@ -81,7 +86,7 @@ def _default_value(rdef, utcnow):
         default = YAMS_TO_PY_TYPEMAP[rdef.object.type](default)
     return default
 
-def iter_subjrdef_card_default(eschema, utcnow):
+def _iter_attr_default(eschema, utcnow):
     for rschema in eschema.subject_relations():
         if rschema.meta:
             continue
@@ -89,9 +94,7 @@ def iter_subjrdef_card_default(eschema, utcnow):
             continue
         rdef = eschema.rdef(rschema.type)
         default = _default_value(rdef, utcnow) if rschema.final else None
-        yield (rschema.type,
-               rdef.cardinality.startswith('1'),
-               default)
+        yield rschema.type, default
 
 # variant of session._update_entity_rel_cache_add that gets
 # the entities
@@ -121,21 +124,53 @@ def _update_entity_rel_cache_add(session, entity, rtype, role, targetentity, oth
         _update_entity_rel_cache_add(session, targetentity, rtype, neg_role(role), entity, True)
 
 
+if notcw319:
+    def reserve_eids(session, qty):
+        """ not fast enough (yet) """
+        source = session.repo.system_source
+        if qty == 1:
+            return (source.create_eid(session),)
+        return source.create_eid(session, count=qty)
+else:
+    def reserve_eids(cnx, qty):
+        source = cnx.repo.system_source
+        if qty == 1:
+            yield source.create_eid(cnx)
+        lasteid = source.create_eid(cnx, count=qty)
+        start = lasteid - qty + 1
+        for eid in xrange(start, lasteid + 1):
+            yield eid
+
+
 class FlushController(object):
     hooksrunnerclass = HooksRunner
     loggername = 'cubicweb'
+    # vectorized hooks are open-coded within the hooks runner
+    vectorized_entity_hooks = ('checkcard_after_add_entity',
+                               'setowner',
+                               'metaattrsinit',
+                               'checkattrconstraint',)
+    deferred_entity_hooks = ('supervising',)
+    vectorized_relation_hooks = ('checkconstraint',)
+    deferred_relation_hooks = ('updateftirel',
+                               'notifyrelationchange')
 
-    def __init__(self, session, schema, disabled_regids,
+    def __init__(self, session,
+                 disabled_regids=(),
                  deferred_entity_hooks=(),
                  deferred_relation_hooks=()):
         self.session = session
-        self.schema = schema
+        self.schema = session.vreg.schema
         self.logger = getLogger(self.loggername)
         self.hooksrunner = self.hooksrunnerclass(self.logger,
                                                  session,
                                                  disabled_regids,
-                                                 deferred_entity_hooks,
-                                                 deferred_relation_hooks)
+                                                 (deferred_entity_hooks +
+                                                  self.vectorized_entity_hooks +
+                                                  self.deferred_entity_hooks),
+                                                 (deferred_relation_hooks +
+                                                  self.vectorized_relation_hooks +
+                                                  self.deferred_relation_hooks))
 
     def insert_relations(self, rtype, fromto):
         session = self.session
@@ -151,16 +186,9 @@ class FlushController(object):
 
         self.hooksrunner.call_rtype_hooks('after_add', rtype, fromto)
 
-    def _reserve_eids(self, qty):
-        """ not fast enough (yet) """
-        assert qty > 0
-        source = self.session.repo.system_source
-        if qty == 1:
-            return (source.create_eid(self.session),)
-        return source.create_eid(self.session, count=qty)
-
     def insert_entities(self, etype, entitiesdicts,
-                        postprocessentity=None):
+                        processentity=None,
+                        processattributes=None):
 
         eschema = self.schema[etype]
         etypeclass = self.session.vreg['etypes'].etype_class(etype)
@@ -175,32 +203,35 @@ class FlushController(object):
         isrelation = []
         isinstanceof = []
         entities = []
-        entitycallbacks = []
         bytesrtypes = set(rschema.type
                           for rschema in eschema.subject_relations()
                           if 'Bytes' in rschema.targets())
 
         utcnow = datetime.utcnow()
 
-        for packedstuff, eid in izip(entitiesdicts,
-                                     self._reserve_eids(len(entitiesdicts))):
+        eidsequence = reserve_eids(self.session, len(entitiesdicts))
+        for attrs_and_callbackdata, eid in izip(entitiesdicts, eidsequence):
 
-            insertattrs = packedstuff[0]
-            callbackdata = packedstuff[1:]
+            insertattrs = attrs_and_callbackdata[0]
 
             # metaattrsinit hook
             insertattrs['creation_date'] = utcnow
             insertattrs['modification_date'] = utcnow
-            insertattrs['cwuri'] = u''
+            if 'cwuri' not in insertattrs:
+                insertattrs['cwuri'] = u''
             insertattrs['eid'] = eid
 
+            if processattributes:
+                processattributes(insertattrs, attrs_and_callbackdata[1])
+
             attributes.append(insertattrs)
             allkeys |= set(insertattrs)
 
             # prepare metadata tables
-            metadata.append({'type': etype, 'eid': eid,
-                             'source': 'system', 'asource': 'system',
-                             'mtime': utcnow})
+            meta = {'type': etype, 'eid': eid, 'asource': 'system'}
+            if notcw319:
+                meta.update({'mtime': utcnow, 'source': 'system'})
+            metadata.append(meta)
             isrelation.append({'eid_from': eid, 'eid_to': etypeid})
             for ancestor in ancestorseid:
                 isinstanceof.append({'eid_from': eid, 'eid_to': ancestor})
@@ -214,11 +245,10 @@ class FlushController(object):
             self.session.set_entity_cache(entity)
 
             entities.append(entity)
-            entitycallbacks.append(packedstuff)
 
         # give a default value to unvalued attributes
         # after this, all attributes are well defined
-        for rtype, required, default in iter_subjrdef_card_default(eschema, utcnow):
+        for rtype, default in _iter_attr_default(eschema, utcnow):
             for attr in attributes:
                 if rtype not in attr:
                     attr[rtype] = default
@@ -226,12 +256,16 @@ class FlushController(object):
         # update the repo.eid_type_source cache
         repo = self.session.repo
         for entity in entities:
-            repo._type_source_cache[entity.eid] = entity.cw_etype, 'system', None, 'system'
-
+            if notcw319:
+                repo._type_source_cache[entity.eid] = entity.cw_etype, 'system', None, 'system'
+            else:
+                repo._type_source_cache[entity.eid] = entity.cw_etype, None, 'system'
         inlinedrtypes = set(rschema.type
                             for rschema in eschema.subject_relations()
                             if rschema.inlined)
 
+        # we compute the smallest possible inlined rtypes set to minimize
+        # the work done by the hook
         irtypes = inlinedrtypes.intersection(allkeys)
         self.hooksrunner.call_etype_hooks('before_add', etype, entities, irtypes)
 
@@ -259,9 +293,9 @@ class FlushController(object):
                 for rtype, data in binary.iteritems():
                     insertattrs[rtype] = data
 
-        if postprocessentity is not None:
-            for entity, callbackdata in izip(entities, entitycallbacks):
-                postprocessentity(entity, *callbackdata)
+        if processentity is not None:
+            for entity, callbackdata in izip(entities, entitiesdicts):
+                processentity(entity, *callbackdata)
 
         self.hooksrunner.call_etype_hooks('after_add', etype, entities, irtypes)
 
@@ -273,16 +307,21 @@ class FlushController(object):
 
         return entities
 
-    def run_deferred_hooks(self, errors, target):
-        self.logger.info('running deferred hooks')
+    def run_deferred_hooks(self, errors):
+        """Run vectorized hooks and pass deferred hooks to a worker task.
+        This must be called explicitly before the end of the transaction.
+        """
+        self.logger.info('running vectorized hooks')
         session = self.session
         schema = session.vreg.schema
         # we either run a 'vectorized' version of these or
         # we get a fresh session^Wtransaction to run this stuff
         # in the context of a worker task
+        vectorized_regids = set(self.vectorized_relation_hooks) | set(self.vectorized_entity_hooks)
         deferred_entity_hooks = []
         deferred_relation_hooks = []
 
+        # Handle vectorized hooks
         # entity hooks
         _ = self.session._
         for regid, entities_by_etype in self.hooksrunner.deferred_entity_hooks.iteritems():
@@ -359,7 +398,8 @@ class FlushController(object):
                 for etype, entities in entities_by_etype.iteritems():
                     for entity in entities:
                         by_etype[etype].append((entity.eid, entity.cw_attr_cache))
-                deferred_entity_hooks.append((regid, by_etype))
+                if regid not in vectorized_regids:
+                    deferred_entity_hooks.append((regid, by_etype))
 
 
         # relation hooks
@@ -389,18 +429,19 @@ class FlushController(object):
             else:
                 if not relations:
                     continue
-                deferred_relation_hooks.append((regid, relations))
+                if regid not in vectorized_regids:
+                    deferred_relation_hooks.append((regid, relations))
 
         if deferred_entity_hooks or deferred_relation_hooks:
             self.logger.info('saving info for %s entity hooks', len(deferred_entity_hooks))
             self.logger.info('saving info for %s relation hooks', len(deferred_relation_hooks))
-            session.create_entity('CWWorkerTask',
-                                  operation=u'run-deferred-hooks',
-                                  target=target,
-                                  on_behalf=session.user,
-                                  deferred_hooks=Binary(dumps((deferred_entity_hooks,
-                                                               deferred_relation_hooks))))
-        self.logger.info('/running deferred hooks')
+            with session.deny_all_hooks_but('metadata', 'workflow'):
+                task = session.create_entity('CWWorkerTask',
+                                             operation=u'run-deferred-hooks',
+                                             deferred_hooks=Binary(dumps((deferred_entity_hooks,
+                                                                          deferred_relation_hooks))))
+                self.logger.info('scheduling task %s to run deferrd hooks', task.eid)
+        self.logger.info('/running vectorized hooks')
 
 def contiguousboundaries(eids):
     """
@@ -452,15 +493,14 @@ def newsession(self, user):
     finally:
         session.close()
 
-class DeferredHooksRunner(EntityAdapter):
+class DeferredHooksRunner(Performer):
     __regid__ = 'run-deferred-hooks'
-    __select__ = is_instance('CWWorkerTask')
 
     def abort_task(self, session, task, error):
         pass
 
     def perform_task(self, session, task):
-        user = task.on_behalf[0]
+        user = task.created_by[0]
         with newsession(session, user) as session:
             entity_hooks, relation_hooks = loads(task.deferred_hooks.getvalue())
             try:
@@ -502,14 +542,16 @@ class DeferredHooksRunner(EntityAdapter):
                     entities.append(entity)
 
                 if hookregid == '__pseudo_entity_fti__':
-                    self.info('%s: fti for %s entities', etype, len(eid_plus_caches))
+                    if server.DEBUG & server.DBG_HOOKS:
+                        print '%s: fti for %s entities' % (etype, len(eid_plus_caches))
                     for entity in entities:
                         entity.complete(entity.e_schema.indexable_attributes())
                         source.index_entity(session, entity=entity)
                     continue
 
                 hookclass, events = self._fetch_hook(hookregid, 'entity')
-                self.info('entity hooks: %s %s (%s)', etype, hookregid, len(entities))
+                if server.DEBUG & server.DBG_HOOKS:
+                    print 'entity hooks: %s %s (%s)' %(etype, hookregid, len(entities))
                 for entity in entities:
                     assert entity.cw_etype == etype
                     with self._cw.security_enabled(read=False, write=False):
@@ -523,12 +565,11 @@ class DeferredHooksRunner(EntityAdapter):
         for hookregid, fromto_by_rtype in relation_hooks:
             hookclass, events = self._fetch_hook(hookregid, 'relation')
             for rtype, fromto in fromto_by_rtype.iteritems():
-                self.info('relation hooks: %s %s (%s)', rtype, hookregid, len(fromto))
+                if server.DEBUG & server.DBG_HOOKS:
+                    print 'relation hooks: %s %s (%s)' % (type, hookregid, len(fromto))
                 with session.security_enabled(read=False, write=False):
                     for eidfrom, eidto in fromto:
                         for event in events:
                             hook = hookclass(session, event=event, rtype=rtype,
                                              eidfrom=eidfrom, eidto=eidto)
                             hook()
-
-
